'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var initSqlJs = require('sql.js');
var react = require('react');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var initSqlJs__default = /*#__PURE__*/_interopDefaultLegacy(initSqlJs);

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

/**
 * Create a SQL query from the input schema model.
 *
 * @param schema    Schema for the database
 *
 * @return A tuple with the query string and a list of the tables.
 */
const createQueryFromSchema = (schema) => {
    const { tables, dataBuffer } = schema;
    const cautionMode = !!dataBuffer;
    const tablesList = Object.keys(tables);
    const createTable = cautionMode
        ? "CREATE TABLE IF NOT EXISTS"
        : "CREATE TABLE";
    let query = "";
    tablesList.forEach(table => {
        const { fields, values } = tables[table];
        const fieldsPart = Object.keys(fields).reduce((acc, field, idx, keys) => {
            return acc + `${field} ${fields[field]}` + (idx < keys.length - 1 ? ', ' : '');
        }, '');
        const insertPart = values
            ? values.map((data) => {
                const dataPart = Object.keys(data).reduce((acc, key, idx, keys) => {
                    const value = typeof data[key] === "string"
                        ? `'${data[key]}'`
                        : data[key].asFunc
                            ? `${data[key].value}`
                            : `${data[key]}`;
                    return acc + (value) + ((idx < keys.length - 1 ? ', ' : ''));
                }, '');
                return `INSERT INTO ${table} VALUES (${dataPart});`;
            }).join('')
            : [];
        query += `${createTable} ${table} (${fieldsPart});${insertPart}`;
    });
    return [query, tablesList];
};

/**
 * Returns an array with uniq values of the input array.
 *
 * @param array Array with duplicated elements
 */
const uniq = (array) => {
    const final = [];
    const cache = {};
    array.forEach(n => {
        if (!cache[n]) {
            cache[n] = true;
            final.push(n);
        }
    });
    return final;
};

/**
 * Module with helper functions to execute Regular Expressions
 */
/**
 * Returns the result of the String.match method against the RegEx:
 * `(?<=((FROM[ \n\r]+)|(JOIN[ \n\r]+)|(APPLY[ \n\r]+)))(${tables.join('|')})`
 *
 * @param query     Query to be tested
 * @param tables    List of database tables
 * @return the result of query.match method
 */
const readingQueryMatch = (query, tables) => {
    const readRegEx = new RegExp(`(?<=((FROM[ \n\r]+)|(JOIN[ \n\r]+)|(APPLY[ \n\r]+)))(${tables.join('|')})`, 'g');
    return query.match(readRegEx);
};
/**
 * Returns the result of the String.match method against the RegEx:
 * `^(INSERT INTO).+(?<table>${tables.join('|')})`
 *
 * @param query     Query to be tested
 * @param tables    List of database tables
 * @return the result of query.match method
 */
const writingQueryMatch = (query, tables) => {
    const readRegEx = new RegExp(`^(INSERT INTO).+(?<table>${tables.join('|')})`);
    return query.match(readRegEx);
};
/**
 * Returns the table on which the query was SELECTing data from.
 *
 * @param query     Query to be tested
 * @param tables    List of database tables
 * @return The tables names, or null
 */
const tableFromReadQuery = (query, tables) => {
    const regExResult = readingQueryMatch(query, tables);
    if (regExResult) {
        return uniq(regExResult);
    }
    return null;
};
/**
 * Returns the table on which the query was INSERTing data to.
 *
 * @param query     Query to be tested
 * @param tables    List of database tables
 * @return The table name, or null
 */
const tableFromWritingQuery = (query, tables) => {
    var _a;
    const regExRes = writingQueryMatch(query, tables);
    const table = (_a = regExRes === null || regExRes === void 0 ? void 0 : regExRes.groups) === null || _a === void 0 ? void 0 : _a.table;
    if (table) {
        return table;
    }
    return null;
};

/**
 * Object that holds the database instance, and provides basic instance
 * management.
 */
const database = {
    instance: null,
    setInstance: function (db) {
        this.instance = db;
    },
    getInstance: function () {
        return this.instance;
    },
    destroy: function () {
        var _a;
        (_a = this.instance) === null || _a === void 0 ? void 0 : _a.close();
        this.instance = null;
    }
};
/**
 * Getter interface to the database.
 */
const getDatabase = () => database.getInstance();
/**
 * Public interface to destroy the database.
 * It executes the db.close() function, and remove it from the singleton holder.
 */
const destroyDatabase = () => {
    database.destroy();
};
/**
 * List of the tables of the database.
 * Used to identify the Read/Write queries.
 */
const tables = [];
/**
 * Object holding the weakRefs "listeners".
 */
const GlobalRecorder = {};
/**
 * Create SQLite instance using sql.js, from the input 'path' to the wasm file,
 * and schema definition.
 *
 *
 * @param path      Path to the wasm file
 * @param schema    Schema definition to create the database
 * @param functions Functions map to be added to the database
 * @return database or undefined
 */
const createSQL = (path, schema, functions) => __awaiter(void 0, void 0, void 0, function* () {
    try {
        const buffer = schema.dataBuffer
            ? new Uint8Array(schema.dataBuffer)
            : undefined;
        const SQL = yield initSqlJs__default["default"]({
            locateFile: () => path
        });
        const db = new SQL.Database(buffer);
        database.setInstance(db);
        if (functions) {
            Object.keys(functions).forEach(name => {
                db.create_function(name, functions[name]);
            });
        }
        const [query, tablesList] = createQueryFromSchema(schema);
        tables.push(...tablesList);
        db.exec(query);
        return db;
    }
    catch (e) {
        console.error(e);
    }
    return;
});
/**
 * Registers the input 'updateState' function in the 'recorder' object
 * for the related table.
 *
 * - The table's name is selected from the query;
 * - The table, and/or the 'updateState' function, are added to the Recorder
 *
 * @param updateState   Function to be recorder as Actuator
 * @param query         SQL query that was executed
 * @param tables        List of tables of the database
 * @param recorder      Recorder object containing the weakRefs
 * @return void
 */
const registerQueryListeners = (updateState, query, tables, recorder) => {
    const foundTables = tableFromReadQuery(query, tables);
    if (foundTables) {
        foundTables.forEach(table => {
            if (!recorder[table]) {
                recorder[table] = [];
            }
            recorder[table].push(new WeakRef(updateState));
        });
    }
};
/**
 * Execute a call to all actuators registered to a certain table.
 *
 * - The table used for the query is selected;
 * - A timestamp is generated;
 * - All actuators, related to the selected table, are executed
 *   asynchronously with the timestamp.
 *
 * @param query     SQL query that was executed
 * @param tables    List of tables of the database
 * @param recorder  Recorder object containing the weakRefs
 * @return void
 */
const triggerActuators = (query, tables, recorder) => {
    const table = tableFromWritingQuery(query, tables);
    if (table) {
        const timestamp = Date.now();
        if (!recorder[table]) {
            return;
        }
        new Promise((resolve, reject) => {
            try {
                recorder[table].forEach(weakRef => {
                    const fnRef = weakRef.deref();
                    fnRef && fnRef(timestamp);
                });
                resolve(1);
            }
            catch (err) {
                console.error('UPDATE ERROR: ', err);
                reject();
            }
        });
    }
};
/**
 * Execute the input 'query' with its 'params' in the database instance
 * using the 'exec' interface.
 *
 * @param query     SQL query to be executed
 * @param params    The BindParams to be used in the exec method.
 * @return the database result or undefined.
 */
const executeQuery = (query, params) => {
    let db = getDatabase();
    if (!db) {
        console.error('SQL-lite instance not initiated! \nRun createSQL function to initialize the service.');
        return;
    }
    try {
        return db.exec(query, params);
        // @ts-ignore
    }
    catch (error) {
        console.error(`SQL-lite Error: ${error.message}`);
        return;
    }
};
/**
 * Executes a SQL query for SELECT purpose, and returns its result.
 *
 * - The updateStateFn is register among the listeners;
 * - The Query is executed;
 * - The Actuators are triggered;
 * - the query result is returned.
 *
 * @param updateStateFn     The updater function to register in the GlobalRecorder
 * @param query             SQL query with the INSERT command
 * @param params            BindParams for the query
 * @return The result from 'executeQuery' function (Array<QueryExecResult> | undefined)
 */
const queryPipeline = (updateStateFn, query, params) => {
    registerQueryListeners(updateStateFn, query, tables, GlobalRecorder);
    const result = executeQuery(query, params);
    triggerActuators(query, tables, GlobalRecorder);
    return result;
};
/**
 * Executes a SQL query for INSERT purpose, and returns its result.
 *
 * - The query containing the INSERT command is validated;
 * - The Query is executed;
 * - The Actuators are triggered;
 * - the query result is returned.
 *
 * @param query     SQL query with the INSERT command
 * @param params    BindParams for the query
 * @param recorder  Recorder object (GlobalRecorder as default)
 * @return The result from 'executeQuery' function (Array<QueryExecResult> | undefined)
 */
const insertQueryPipeline = (query, params, recorder = GlobalRecorder) => {
    const regExRes = writingQueryMatch(query, tables);
    if (!regExRes) {
        return;
    }
    const result = executeQuery(query, params);
    triggerActuators(query, tables, recorder);
    return result;
};

function useQuery(query, depths) {
    const [updatedAt, setUpdatedAt] = react.useState();
    const writeQueryFn = react.useCallback((query) => {
        insertQueryPipeline(query);
    }, []);
    return {
        result: queryPipeline(setUpdatedAt, query),
        writeQueryFn
    };
}

/**
 * Hook that provide write-only access to the database
 *
 */
function useInsert() {
    return react.useCallback((query) => {
        insertQueryPipeline(query);
    }, []);
}

/**
 * Hook that provide read-only access to the database
 *
 * @param query     Query to be executed
 * @param depths
 */
function useSelect(query, depths) {
    const [updatedAt, setUpdatedAt] = react.useState();
    return queryPipeline(setUpdatedAt, query);
}

exports["default"] = createSQL;
exports.destroyDatabase = destroyDatabase;
exports.useInsert = useInsert;
exports.useQuery = useQuery;
exports.useSelect = useSelect;
//# sourceMappingURL=index.js.map
