import { Database, BindParams, QueryExecResult } from 'sql.js';
import { DatabaseHolder, Functions, Recorder, Schema } from "../../types";
/**
 * Object that holds the database instance, and provides basic instance
 * management.
 */
export declare const database: DatabaseHolder;
/**
 * Getter interface to the database.
 */
export declare const getDatabase: () => Database | null;
/**
 * Public interface to destroy the database.
 * It executes the db.close() function, and remove it from the singleton holder.
 */
export declare const destroyDatabase: () => void;
/**
 * Create SQLite instance using sql.js, from the input 'path' to the wasm file,
 * and schema definition.
 *
 *
 * @param path      Path to the wasm file
 * @param schema    Schema definition to create the database
 * @param functions Functions map to be added to the database
 * @return database or undefined
 */
export declare const createSQL: (path: string, schema: Schema, functions?: Functions | undefined) => Promise<Database | undefined>;
/**
 * Registers the input 'updateState' function in the 'recorder' object
 * for the related table.
 *
 * - The table's name is selected from the query;
 * - The table, and/or the 'updateState' function, are added to the Recorder
 *
 * @param updateState   Function to be recorder as Actuator
 * @param query         SQL query that was executed
 * @param tables        List of tables of the database
 * @param recorder      Recorder object containing the weakRefs
 * @return void
 */
export declare const registerQueryListeners: (updateState: (time: number) => void, query: string, tables: Array<string>, recorder: Recorder) => void;
/**
 * Execute a call to all actuators registered to a certain table.
 *
 * - The table used for the query is selected;
 * - A timestamp is generated;
 * - All actuators, related to the selected table, are executed
 *   asynchronously with the timestamp.
 *
 * @param query     SQL query that was executed
 * @param tables    List of tables of the database
 * @param recorder  Recorder object containing the weakRefs
 * @return void
 */
export declare const triggerActuators: (query: string, tables: Array<string>, recorder: Recorder) => void;
/**
 * Execute the input 'query' with its 'params' in the database instance
 * using the 'exec' interface.
 *
 * @param query     SQL query to be executed
 * @param params    The BindParams to be used in the exec method.
 * @return the database result or undefined.
 */
export declare const executeQuery: (query: string, params?: BindParams | undefined) => Array<QueryExecResult> | undefined;
/**
 * Executes a SQL query for SELECT purpose, and returns its result.
 *
 * - The updateStateFn is register among the listeners;
 * - The Query is executed;
 * - The Actuators are triggered;
 * - the query result is returned.
 *
 * @param updateStateFn     The updater function to register in the GlobalRecorder
 * @param query             SQL query with the INSERT command
 * @param params            BindParams for the query
 * @return The result from 'executeQuery' function (Array<QueryExecResult> | undefined)
 */
export declare const queryPipeline: (updateStateFn: (time: number) => void, query: string, params?: BindParams | undefined) => Array<QueryExecResult> | undefined;
/**
 * Executes a SQL query for INSERT purpose, and returns its result.
 *
 * - The query containing the INSERT command is validated;
 * - The Query is executed;
 * - The Actuators are triggered;
 * - the query result is returned.
 *
 * @param query     SQL query with the INSERT command
 * @param params    BindParams for the query
 * @param recorder  Recorder object (GlobalRecorder as default)
 * @return The result from 'executeQuery' function (Array<QueryExecResult> | undefined)
 */
export declare const insertQueryPipeline: (query: string, params?: BindParams | undefined, recorder?: Recorder) => Array<QueryExecResult> | undefined;
declare const _default: {
    createSQL: (path: string, schema: Schema, functions?: Functions | undefined) => Promise<Database | undefined>;
    destroyDatabase: () => void;
    queryPipeline: (updateStateFn: (time: number) => void, query: string, params?: BindParams | undefined) => QueryExecResult[] | undefined;
    insertQueryPipeline: (query: string, params?: BindParams | undefined, recorder?: Recorder) => QueryExecResult[] | undefined;
};
/**
 * Functions to be exposed to the API.
 */
export default _default;
